<?php

namespace Drupal\nested_entity_reference_formatter\Plugin\Field\FieldFormatter;

use Drupal\Component\Plugin\Factory\DefaultFactory;
use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Field\FormatterPluginManager;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Field\Plugin\Field\FieldFormatter\EntityReferenceFormatterBase;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityDisplayBase;

/**
 * Plugin implementation of the 'nested entity reference' formatter.
 *
 * @FieldFormatter(
 *   id = "nested_entity_reference",
 *   label = @Translation("Nested Entity Reference"),
 *   description = @Translation("Display the nested entity reference formmater."),
 *   field_types = {
 *     "entity_reference",
 *     "entity_reference_revisions"
 *   }
 * )
 */
class NestedEntityReferenceFormatter extends EntityReferenceFormatterBase {

  /**
   * The number of times this formatter allows rendering the same entity.
   *
   * @var int
   */
  public const RECURSIVE_RENDER_LIMIT = 20;

  /**
   * The logger factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity display repository.
   *
   * @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface
   */
  protected $entityDisplayRepository;

  /**
   * The entity type bundle info service.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $entityTypeBundleInfo;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The formatter manager.
   *
   * @var \Drupal\Core\Field\FormatterPluginManager
   */
  protected $formatterManager;


  /**
   * An array of counters for the recursive rendering protection.
   *
   * Each counter takes into account all the relevant information about the
   * field and the referenced entity that is being rendered.
   *
   * @var array
   *
   * @see \Drupal\Core\Field\Plugin\Field\FieldFormatter\EntityReferenceEntityFormatter::viewElements()
   */
  protected static $recursiveRenderDepth = [];

  /**
   * Constructs a EntityReferenceEntityFormatter instance.
   *
   * @param string $plugin_id
   *   The plugin_id for the formatter.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The definition of the field to which the formatter is associated.
   * @param array $settings
   *   The formatter settings.
   * @param string $label
   *   The formatter label display setting.
   * @param string $view_mode
   *   The view mode.
   * @param array $third_party_settings
   *   Any third party settings settings.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityDisplayRepositoryInterface $entity_display_repository
   *   The entity display repository.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entity_type_bundle_info
   *   The entity type bundle info.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\Core\Field\FormatterPluginManager $formatter_manager
   *   The formatter manager.
   */
  public function __construct(string $plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, string $label, string $view_mode, array $third_party_settings, LoggerChannelFactoryInterface $logger_factory, EntityTypeManagerInterface $entity_type_manager, EntityDisplayRepositoryInterface $entity_display_repository, EntityTypeBundleInfoInterface $entity_type_bundle_info, EntityFieldManagerInterface $entity_field_manager, FormatterPluginManager $formatter_manager) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $label, $view_mode, $third_party_settings);
    $this->loggerFactory = $logger_factory;
    $this->entityTypeManager = $entity_type_manager;
    $this->entityDisplayRepository = $entity_display_repository;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
    $this->entityFieldManager = $entity_field_manager;
    $this->formatterManager = $formatter_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings'],
      $container->get('logger.factory'),
      $container->get('entity_type.manager'),
      $container->get('entity_display.repository'),
      $container->get('entity_type.bundle.info'),
      $container->get('entity_field.manager'),
      $container->get('plugin.manager.field.formatter')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return ['reference_fields' => []] + parent::defaultSettings();
  }

  /**
   * Returns an array of options for a field.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getPluginOptions(): array {
    $target_bundles = $this->getTargetBundles();
    $target_type = $this->getFieldSetting('target_type');
    $options = [];
    foreach ($target_bundles as $bundle_name => $bundle_label) {
      $field_definitions = $this->entityFieldManager->getFieldDefinitions($target_type, $bundle_name);
      foreach ($field_definitions as $field_definition) {
        if (!isset($options[$bundle_name])) {
          $options[$bundle_name] = [];
        }
        $options[$bundle_name][$field_definition->getName()] = $field_definition->getLabel();
      }
    }

    return $options;
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException|
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $target_bundles = $this->getTargetBundles();
    $options = $this->getPluginOptions();

    $default_fields = $this->getSetting('reference_fields') ?? [];

    $elements['reference_fields'] = [
      '#open' => FALSE,
      '#title' => $this->t('Fields'),
      '#tree' => TRUE,
      '#required' => TRUE,
    ];

    $field_name = $this->fieldDefinition->getName();

    foreach ($options as $bundle => $fields) {
      $css_id = Html::getUniqueId('edit-' . $bundle . '-wrapper');

      $elements['reference_fields'][$bundle] = [
        '#type' => 'fieldset',
        '#title' => $target_bundles[$bundle],
        '#prefix' => '<div id="' . $css_id . '">',
        '#suffix' => '</div>',
      ];

      $parents = [
        'fields',
        $field_name,
        'settings_edit_form',
        'settings',
        'reference_fields',
        $bundle,
        'field',
      ];

      $reference_field = $form_state->getValue($parents);

      if (!$reference_field) {
        $reference_field = $default_fields[$bundle]['field'] ?? '';
      }

      array_unshift($fields, $this->t('None'));

      $elements['reference_fields'][$bundle]['field'] = [
        '#title' => $this->t('Field'),
        '#type' => 'select',
        '#default_value' => $reference_field,
        '#options' => $fields,
        '#ajax' => [
          'callback' => [$this, 'fieldAjaxCallback'],
          'wrapper' => $css_id,
          'effect' => 'fade',
          'field' => $reference_field,
        ],
      ];

      if ($reference_field) {
        $settings = $default_fields[$bundle]['formatter'] ?? [];
        $this->getFormatterForm($elements['reference_fields'][$bundle], $bundle, $reference_field, $settings);
      }
    }

    return $elements;
  }

  /**
   * Render API callback: gets the field elements.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return array
   *   Form items.
   */
  public function fieldAjaxCallback(array $form, FormStateInterface $form_state) {
    $parents = $form_state->getTriggeringElement()['#array_parents'];
    array_pop($parents);
    return NestedArray::getValue($form, $parents);
  }

  /**
   * Returns an array of applicable formatter options for a field.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition.
   *
   * @return array
   *   An array of applicable formatter options.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @see \Drupal\field_ui\Form\EntityDisplayFormBase::getApplicablePluginOptions()
   */
  protected function getApplicablePluginOptions(FieldDefinitionInterface $field_definition): array {
    $options = $this->formatterManager->getOptions($field_definition->getType());
    $applicable_options = [];
    foreach ($options as $option => $label) {
      $plugin_class = DefaultFactory::getPluginClass($option, $this->formatterManager->getDefinition($option));
      if ($plugin_class::isApplicable($field_definition)) {
        $applicable_options[$option] = $label;
      }
    }
    return $applicable_options;
  }

  /**
   * Shows formatter settings form for the selected field.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param string $bundle
   *   The field bundle.
   * @param string $field_name
   *   The field name.
   * @param array $settings
   *   The formatter settings.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getFormatterForm(array &$form, string $bundle, string $field_name, array $settings) {
    $target_type = $this->getFieldSetting('target_type');
    $definitions = $this->entityFieldManager->getFieldDefinitions($target_type, $bundle);
    /** @var \Drupal\Core\Field\FieldDefinitionInterface $field_config */
    $field_definition = $definitions[$field_name];

    $form['formatter'] = [
      '#tree' => TRUE,
      '#process' => [
        [$this, 'formatterSettingsProcessCallback'],
      ],
      '#bundle' => $bundle,
      '#field' => $field_name,
    ];
    $form['formatter']['label'] = [
      '#type' => 'value',
      '#value' => 'hidden',
    ];

    // Generate a unique wrapper HTML ID.
    $wrapper_id = Html::getUniqueId('formatter-settings-wrapper');

    // @todo remove validated, need to find a better way.
    $form['formatter']['type'] = [
      '#type' => 'select',
      '#title' => $this->t('Formatter'),
      '#options' => $this->getApplicablePluginOptions($field_definition),
      '#required' => TRUE,
      '#validated' => TRUE,
      '#default_value' => $settings['type'] ?? [],
      '#ajax' => [
        'callback' => [$this, 'formatterSettingsAjaxCallback'],
        'wrapper' => $wrapper_id,
      ],
    ];

    // Add the formatter settings to the form via AJAX.
    $form['formatter']['settings_wrapper'] = [
      '#prefix' => '<div id="' . $wrapper_id . '">',
      '#suffix' => '</div>',
    ];

  }

  /**
   * Render API callback: builds the formatter settings elements.
   *
   * @param array $element
   *   The form element to process. Properties used:.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $complete_form
   *   The complete form structure.
   *
   * @return array
   *   The form element.
   */
  public function formatterSettingsProcessCallback(array &$element, FormStateInterface $form_state, array &$complete_form) {
    if ($formatter = $this->getFormatter($element, $form_state)) {
      $element['settings_wrapper']['settings'] = $formatter->settingsForm($complete_form, $form_state);
      $element['settings_wrapper']['settings']['#parents'] = array_merge($element['#parents'], ['settings']);
    }
    return $element;
  }

  /**
   * Render API callback: gets formatter setting element.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return array
   *   Form items.
   */
  public function formatterSettingsAjaxCallback(array $form, FormStateInterface $form_state) {
    $parents = $form_state->getTriggeringElement()['#array_parents'];
    array_pop($parents);
    $parents = array_merge($parents, ['settings_wrapper']);
    return NestedArray::getValue($form, $parents);
  }

  /**
   * Gets the formatter object.
   *
   * @param array $element
   *   The element representing the formatter.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return \Drupal\Core\Field\FormatterInterface|null
   *   The formatter object.
   */
  protected function getFormatter(array $element, FormStateInterface $form_state) {
    $parents = $element['#parents'];
    $bundle = $element['#bundle'];
    $field_name = $element['#field'];
    // Use the processed values, if available.
    $configuration = NestedArray::getValue($form_state->getValues(), $parents);
    if (empty($configuration)) {
      // Next check the raw user input.
      $configuration = NestedArray::getValue($form_state->getUserInput(), $parents);
    }
    if (empty($configuration)) {
      // If no user input exists, use the default values.
      $default_fields = $this->getSetting('reference_fields') ?? [];
      $configuration = $default_fields[$bundle]['formatter'] ?? [];
    }

    $target_type = $this->getFieldSetting('target_type');
    $definitions = $this->entityFieldManager->getFieldDefinitions($target_type, $bundle);

    // Set options.
    $options = [
      'configuration' => $configuration,
      'field_definition' => $definitions[$field_name],
      'view_mode' => EntityDisplayBase::CUSTOM_MODE,
      'prepare' => TRUE,
    ];

    // In some cases, fields do not have any 'default_formatter'
    // formatter as example UUID.
    if (!empty($configuration) && empty($configuration['type'])) {
      return NULL;
    }

    return $this->formatterManager->getInstance($options);
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary(): array {
    $summary = [];
    $settings = $this->getSetting('reference_fields') ?? [];
    $bundles = $this->getTargetBundles();
    foreach ($settings as $bundle => $bundle_settings) {
      $summary[] = [
        '#type' => 'markup',
        '#markup' => '<strong>' . $bundles[$bundle] . '</strong>',
      ];
      $target_type = $this->getFieldSetting('target_type');
      $field_name = $bundle_settings['field'] ?? '';
      if ($field_name) {
        $configuration = $settings[$bundle]['formatter'] ?? [];
        $definitions = $this->entityFieldManager->getFieldDefinitions($target_type, $bundle);

        if (!empty($configuration) && empty($configuration['type'])) {
          continue;
        }

        $formatter = $this->formatterManager->getInstance([
          'configuration' => $configuration,
          'field_definition' => $definitions[$field_name],
          'view_mode' => EntityDisplayBase::CUSTOM_MODE,
          'prepare' => TRUE,
        ]);

        if ($formatter) {
          $formatter_summary = $formatter->settingsSummary();
          $summary = array_merge($summary, $formatter_summary);
        }
      }
      else {
        $summary[] = $this->t('None');
      }
    }
    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $settings = $this->getSetting('reference_fields');
    $elements = [];

    foreach ($this->getEntitiesToView($items, $langcode) as $delta => $entity) {
      // Due to render caching and delayed calls, the viewElements() method
      // will be called later in the rendering process through a '#pre_render'
      // callback, so we need to generate a counter that takes into account
      // all the relevant information about this field and the referenced
      // entity that is being rendered.
      $recursive_render_id = $items->getFieldDefinition()
        ->getTargetEntityTypeId()
        . $items->getFieldDefinition()->getTargetBundle()
        . $items->getName()
        // We include the referencing entity, so we can render default images
        // without hitting recursive protections.
        . $items->getEntity()->id()
        . $entity->getEntityTypeId()
        . $entity->id();

      if (isset(static::$recursiveRenderDepth[$recursive_render_id])) {
        static::$recursiveRenderDepth[$recursive_render_id]++;
      }
      else {
        static::$recursiveRenderDepth[$recursive_render_id] = 1;
      }

      // Protect ourselves from recursive rendering.
      if (static::$recursiveRenderDepth[$recursive_render_id] > static::RECURSIVE_RENDER_LIMIT) {
        $this->loggerFactory->get('entity')
          ->error('Recursive rendering detected when rendering entity %entity_type: %entity_id, using the %field_name field on the %bundle_name bundle. Aborting rendering.', [
            '%entity_type' => $entity->getEntityTypeId(),
            '%entity_id' => $entity->id(),
            '%field_name' => $items->getName(),
            '%bundle_name' => $items->getFieldDefinition()->getTargetBundle(),
          ]);
        return $elements;
      }
      if (empty($settings[$entity->bundle()]['field'])) {
        continue;
      }
      $bundle_settings = $settings[$entity->bundle()];
      if (!empty($entity->{$bundle_settings['field']})) {
        $elements[$delta] = $entity->{$bundle_settings['field']}->view($bundle_settings['formatter']);
      }
    }

    return $elements;
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(FieldDefinitionInterface $field_definition) {
    // This formatter is only available for entity types that have a view
    // builder.
    $target_type = $field_definition->getFieldStorageDefinition()
      ->getSetting('target_type');
    return \Drupal::entityTypeManager()
      ->getDefinition($target_type)
      ->hasViewBuilderClass();
  }

  /**
   * Provides a list of possible target bundles for this field.
   *
   * @return array
   *   The list of target bundles.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getTargetBundles(): array {
    $target_type = $this->getFieldSetting('target_type');
    $handler_settings = $this->getFieldSetting('handler_settings');
    $entity_type = $this->entityTypeManager->getDefinition($target_type);
    $target_bundles = [];
    if (isset($handler_settings['target_bundles']) && $handler_settings['target_bundles'] === []) {
      // Error.
      return $target_bundles;
    }
    if (empty($handler_settings['target_bundles'])) {
      $bundles = $this->entityTypeBundleInfo->getBundleInfo($target_type);
      if ($entity_type->hasKey('bundle')) {
        foreach ($bundles as $bundle_name => $bundle_info) {
          $target_bundles[$bundle_name] = $bundle_info['label'];
        }
      }
    }
    else {
      $bundles = $this->entityTypeBundleInfo->getBundleInfo($target_type);
      foreach ($handler_settings['target_bundles'] as $bundle_name) {
        $target_bundles[$bundle_name] = $bundles[$bundle_name]['label'];
      }
    }
    return $target_bundles;
  }

}
